/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "Channel4Base.h"
#include <boost/format.hpp>
#include <time.h>
#include <stdio.h>
#ifndef _MSC_VER
#include <sys/ioctl.h>
#endif
#include <memory.h>

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param id Nummer der Kanalklasse fuer z.B. Trace.
 */
Channel4Base::Channel4Base( int id ) 
{
    // Kanalnummer zuweisen
    _id = id;

    // Membervars initialisieren
    _type    = Undefined;
    _comm    = INVALID_HANDLE_VALUE;
    _socket  = INVALID_SOCKET;
    _value   = 0;
    _timeout = 0;
#ifndef _MSC_VER
    memset( &_flock, 0, sizeof(struct flock) );
    memset( &_termio, 0, sizeof(struct termio) );
#endif

    _endSignal = 0;
    _initSignal = 0;
    _communicationServerThreadState = Down;
    _communicationServerThread = 0;
    // Socket fuer Listensocket.
    _listenSocket = INVALID_SOCKET;

    memset( &_serverThreadData, 0, sizeof(_serverThreadData) );

    // Statistik
    _writtenSize = 0;
    _amountSize = 0;

    // Tracemeldungen
    _trace  = FALSE;
    // Tracefilename
    _traceFileName = boost::str( boost::format( "DFComDLL_%1%.log" ) % _id );

#ifdef _DEBUG
    // Nur im Debugmodus standardmaessig aktiv.
    setTrace();
#endif
}

// ##################################################################
/**
 * @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
 *
 */
Channel4Base::~Channel4Base() 
{
    // Schliessen der Schnittstelle, wenn nicht bereits.
    close();
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

// ##################################################################
/**
 * @brief Verwirft anstehende Zeichen der Schnittstelle.
 */
void Channel4Base::flush()
{
    trace( "%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

    unsigned char buffer[0x100];
    int readCount = 0xFF;

    // eventuell anstehende Datenpakete verwerfen
    int i;
    for ( i = 0; (i < 5) && readCount; ++i )
    {
        if ( readData( buffer, sizeof(buffer) - 1, readCount ) != IV_ERROR_SUCCESS )
        {	
            trace( "%s(%i):INFO(2): %s() [ Abbruch nach %i durchläufen. ]\n", __FILE__, __LINE__, __func__, i );
#ifdef _DEBUG
            int z; 
            for ( z = 0; z < readCount; ++z ) 
                trace( "%02X ", buffer[z] );

            trace( "\n" );
#endif
            break;
        }

        trace( "%s(%i):INFO(2): %s() [ Verworfene Anzahl Zeichen (%i Byte) ]\n", __FILE__, __LINE__, __func__, readCount );
#ifdef _DEBUG
        int z; 
        for ( z = 0; z < readCount; ++z ) 
            trace( "%02X ", buffer[z] );

        trace( "\n" );
#endif
    }
}

// ##################################################################
/**
 * @brief Hauptlogfunktion der Kanalklassen.
 */
void Channel4Base::trace( const char *format, ... )
{
    va_list ap;
    char szTime[14];
#ifdef _MSC_VER
    char szID[20];
#endif
    char buf[0x800];
    char *pos;
    FILE *fp;
    unsigned short msec = 0;

    if ( isTrace() )
    {
        if (/*strstr(format, "INFO(1)") != NULL ||
            strstr(format, "INFO(2)") != NULL ||*/
            strstr(format, "INFO(3)") != NULL || /*
          strstr(format, "WARNING(1)") != NULL ||
          strstr(format, "WARNING(2)") != NULL ||*/
          strstr(format, "WARNING(3)") != NULL /*||
          strstr(format, "ERROR(1)") != NULL ||
          strstr(format, "ERROR(2)") != NULL ||
          strstr(format, "ERROR(3)") != NULL*/ )
        { 
            return;
        }

        va_start( ap, format );
        vsprintf( buf, format, ap );
        va_end( ap );
        // Uhrzeit des Logeintrages
        time_t t = time( NULL );
        tm *ltm = localtime( &t );
#ifdef _MSC_VER
        SYSTEMTIME st;
        GetLocalTime( &st );
        msec = st.wMilliseconds;
#endif
        sprintf( szTime, ":%02i:%02i:%02i.%03i", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, msec );
        if ( (pos = (char*)strstr(buf, ":INFO(")) != NULL || 
            (pos = (char*)strstr(buf, ":WARNING(")) != NULL || 
            (pos = (char*)strstr(buf, ":ERROR(")) != NULL )
        {
                // Uhrzeit einfuegen
                memmove( &buf[(pos - buf) + 13], pos, strlen( pos ) + 1 );
                strncpy( pos, szTime, 13 );
#ifdef _MSC_VER
                // ThreadID einfuegen
                memset( szID, 0, sizeof(szID) );
                sprintf( szID, ":(%i, %u)", getId(), static_cast<unsigned>( GetCurrentThreadId() ) );
                memmove( &buf[(pos - buf) + strlen( szID )], pos, strlen( pos ) + 1 );
                strncpy( pos, szID, strlen( szID ) );
#endif
        }

#ifndef _MSC_VER
        if ( buf[strlen(buf) - 1] == '\n' )
            strcpy( &buf[strlen(buf) - 1], "\r\n" );
#endif

        if ( (fp = fopen(_traceFileName.data(), "a")) == (FILE*)0 )
            return;

        fwrite( buf, 1, strlen( buf ), fp );
        fclose( fp );

#ifdef _DEBUG
#ifdef _MSC_VER
        OutputDebugStringA( buf );
#endif
#endif
    }
}

// ##################################################################
/**
 * @brief Oeffnen der Kommunikationsschnittstelle.
 */
int Channel4Base::open( enumType type, const std::string &string, int val, int timeout )
{
    trace( "%s(%i):INFO(3): %s( %i, %p, %i, %i )\n", __FILE__, __LINE__, __func__, type, string.c_str(), val, timeout );

    int res = IV_ERROR_FAIL;

    switch ( type )
    {
        case Serial:
        
            if ( timeout == 0 )
            {
                // Von einem Standardtimeout 800ms ausgehen.
                if ( _timeout == 0 )
                    timeout = 800;
                else
                    timeout = _timeout; // Zuvor gewaehlte.

            }
            else if ( timeout < 500 )
            { 
                // Mindesttimeout sicherstellen.
                trace( "%s(%i):WARNING(1): %s() [ Zu kleiner Timeout (%i), wurde auf 500 gesetzt. ]\n", __FILE__, __LINE__, __func__, timeout );
                timeout = 500;
            }

            if ( val == 0 )
            {
                // Von einer Standardbaudrate 38400bd ausgehen.
                if ( _value == 0 )
                    val = 38400;
                else 
                    val = _value; // Zuvor gewaehlte.
            }

            if ( openSerial( string, val, timeout ) == IV_ERROR_SUCCESS )
            {
                _type = Serial;
                _value = val;
                _timeout = timeout;
                _commString = string;

                res = IV_ERROR_SUCCESS;
            }
            break;

        case Socket:
            // Mindesttimeout sicherstellen.
            if ( timeout < 2000 )
            {
                trace( "%s(%i):WARNING(1): %s() [ Zu kleiner Timeout (%i), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, timeout );
                timeout = 2000;
            }

            if ( openSocket( string, val, timeout ) == IV_ERROR_SUCCESS )
            {
                _type = Socket;
                _value = val;
                _timeout = timeout;
                _commString = string;

                res = IV_ERROR_SUCCESS;
            }
            break;

       case SocketServer:
            // Mindesttimeout sicherstellen.
            if ( timeout < 2000 )
            {
                trace( "%s(%i):WARNING(1): %s() [ Zu kleiner Timeout (%i), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, timeout );
                timeout = 2000;
            }

            if ( openSocketServer( string, val, timeout ) == IV_ERROR_SUCCESS )
            {
                _type = SocketServer;
                _value = val;
                _timeout = timeout;
                _commString = string;

                res = IV_ERROR_SUCCESS;
            }
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, __func__, type );
            return res;
    }

    return res;
}

// ##################################################################
/**
 * @brief Ermitteln ob die Kommunikationsschnittstelle geoeffnet ist.
 */
int Channel4Base::isOpen()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    return (getType() != Undefined);
}

// ##################################################################
/**
 * @brief Schliessen der Kommunikationsschnittstelle.
 */
int Channel4Base::close()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    switch ( getType() )
    {
        case Undefined:
            return IV_ERROR_SUCCESS;

        case Socket:
            closeSocket();
            break;

        case Serial:
            closeSerial();
            break;

        case SocketServer:
            closeSocketServer();
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, __func__, getType() );
            return IV_ERROR_FAIL;
    }

    _type = Undefined;
    _commString = "";

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Zuweisen einer vorhandenen Kommunikationsschnittstelle.
 * 
 */
int Channel4Base::attachChannel( enumType type, HANDLE val, int timeout )
{
    trace("%s(%i):INFO(3): %s( %i, %i, %i )\n", __FILE__, __LINE__, __func__, type, val, timeout);

    bool wasUndefined = (getType() == Undefined);

    if ( getType() == Undefined
        || (getType() == type && type == Serial && val == _comm)
        || (getType() == type && type == Socket && val == (HANDLE)_socket) )
    {
        switch ( type )
        {
            case Serial:
                _type = Serial;
                _comm = (HANDLE)val;
                _timeout   = timeout;
                break;

            case Socket:
                if ( timeout < 2000 )
                {		
                    trace( "%s(%i):WARNING(1): %s() [ Zu kleiner Timeout (%i), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, __func__, timeout );
                    timeout = 2000;
                }
                _type   = Socket;
                _socket = (SOCKET)val;
                _timeout     = timeout;
                break;

            default:
                trace( "%s(%i):ERROR(1): %s() [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, __func__, type );
                return IV_ERROR_FAIL;
        }

        if ( wasUndefined ) // Kanal Zugewiesen
            _commString = "ATTACH";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%i):ERROR(1): %s() [ Zuweisen des Kanals nicht moeglich, da nicht Undefiniert. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;

}

// ##################################################################
/**
 * @brief Freigeben der aktuellen Kommunikationsschnittstelle.
 */
int Channel4Base::detachChannel()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    if ( getType() == Serial || _commString == "ATTACH" )
    {
        _type = Undefined;
        _value = 0;
        _timeout   = 0;
        _comm = INVALID_HANDLE_VALUE;
        _socket = INVALID_SOCKET;

        // Kommunikationsstring loeschen
        _commString = "";

        return IV_ERROR_SUCCESS;
    }

    trace( "%s(%i):ERROR(1): %s() [ Freigabe des Kanals nicht moeglich, da nicht Seriell. ]\n", __FILE__, __LINE__, __func__ );

    return IV_ERROR_FAIL;
}

// ##################################################################
/**
 * @brief Oeffnen einer seriellen Kommunikationsschnittstelle.
 */
int Channel4Base::openSerial( const std::string &comm, int baudRate, int timeout )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i )\n", __FILE__, __LINE__, __func__, comm.c_str(), baudRate, timeout );

    int rc;

    if ( getType() != Undefined )
    {
        // Kanal bereits geoeffnet
        trace( "%s(%i):ERROR(1): %s() [ Kanal bereits geoeffnet. ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

#ifdef _MSC_VER

    if ( (_comm = CreateFileA(
                comm.data(),					// pointer to name of the file 
                (GENERIC_READ | GENERIC_WRITE),	// access (read-write) mode 
                0,								// share mode 
                NULL,							// pointer to security descriptor 
                OPEN_EXISTING,					// how to create 
                0,								// file attributes 
                0 								// handle to file with attributes to copy  
                )) == INVALID_HANDLE_VALUE )
    { 
        trace( "%s(%i):ERROR(1): %s() [ CreateFile fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    errno = 0;

    if ( (rc = ::open( comm.data(), O_RDWR | O_NDELAY )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ open fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    _comm = rc;

    if ( (rc = fcntl( _comm, F_SETFL, O_RDWR )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ fcntl fail ]\n", __FILE__, __LINE__, __func__ );
        return IV_ERROR_FAIL;
    }

    _flock.l_type    = F_WRLCK;  // lock stream for read and write
    _flock.l_whence  = 0;        // lock whole file -> 0,0,0
    _flock.l_start   = 0;
    _flock.l_len     = 0;

    if ( (rc = fcntl( _comm, F_SETLK, &_flock )) < 0 )
    {
        trace( "%s(%i):WARINING(1): %s() [ fcntl (file locking) fail ]\n", __FILE__, __LINE__, __func__ );
        //return IV_ERROR_FAIL;
    }
#endif

    if ( (rc = changeStateSerial( _comm, baudRate, timeout )) != IV_ERROR_SUCCESS )
    {
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Aendern der aktuellen Parameter zur seriellen Kommunikationsschnittstelle.
 */
int Channel4Base::changeStateSerial( HANDLE handle, int baudRate, int timeout )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i )\n", __FILE__, __LINE__, __func__, handle, baudRate, timeout);

#ifdef _MSC_VER
    DCB          dcb;
    COMMTIMEOUTS ctout;
    DWORD        cerr = 0;

    memset( &dcb, 0, sizeof(dcb) );
    dcb.DCBlength = sizeof(dcb);
    dcb.fBinary   = 1;

    dcb.BaudRate  = baudRate;
    dcb.ByteSize  = 8;
    dcb.Parity    = NOPARITY;
    dcb.StopBits  = ONESTOPBIT;

    if ( !SetCommState( handle, &dcb ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ SetCommState fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( handle == _comm )
    {
        // Akutuelle Baudrate anpassen.
        _value = baudRate;
    }

    ctout.ReadIntervalTimeout         = 0;
    ctout.ReadTotalTimeoutMultiplier  = 0;
    ctout.ReadTotalTimeoutConstant    = (DWORD)timeout;
    ctout.WriteTotalTimeoutConstant   = 0;
    ctout.WriteTotalTimeoutMultiplier = 0;

    if ( !SetCommTimeouts( handle, &ctout ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ SetCommTimeouts fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( handle == _comm )
    {
        // Aktuellen Timeout anpassen
        _timeout = timeout;
    }

    // unsolved: fuer modems + RS485 notwendig !?!
    if ( !EscapeCommFunction( handle, SETRTS ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }
    if ( !EscapeCommFunction( handle, SETDTR ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ EscapeCommFunction fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( !ClearCommError( handle, &cerr, NULL ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ ClearCommError fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

#else
    int rc;
    int baudmode, charmode, paricheck, stopmode;
    int termio_cm;

    switch( baudRate )
    {
        case 9600:
            baudmode = B9600;
            break;

        case 19200:
            baudmode = B19200;
            break;

        case 38400:
            baudmode = B38400;
            break;
		case 57600:
			baudmode = B57600;
			break;
		case 115200:
			baudmode = B115200;
			break;
        default:
            return IV_ERROR_FAIL;
    }

    charmode = CS8;
    paricheck = 0;
    stopmode = 0;

    _termio.c_iflag = 0;
    _termio.c_oflag = 0;
    _termio.c_cflag = 0;
    _termio.c_lflag = 0;
    _termio.c_line  = 0;
    memset( (void *)_termio.c_cc, 0, NCCS );
   
    _termio.c_cflag |= HUPCL;        // hangup line when close/exit(DTR of)
    _termio.c_cflag |= CLOCAL;       // local/direct connection
   
    _termio.c_cflag |= baudmode;     // set baudrate
    _termio.c_cflag |= charmode;     // set databits
    _termio.c_cflag |= paricheck;    // set parity
    _termio.c_cflag |= stopmode;     // set stopbits

    _termio.c_cflag |= CREAD;        // enable input queue

    _termio.c_cc[VMIN]  = 1;         // read waiting for 1 char
   
    if ( (rc = ioctl( handle, TCSETA, &_termio )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( handle, TCFLSH, TCIOFLUSH )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( (rc = ioctl( handle, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    if ( !(termio_cm & TIOCM_RTS) )
        termio_cm |= TIOCM_RTS;

    if ( !(termio_cm & TIOCM_DTR) )
        termio_cm |= TIOCM_DTR;

    if ( (rc = ioctl( handle, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        closeSerial();
        return IV_ERROR_FAIL;
    }

    // Handelt es sich um die eigene Schnittstelle?
    if ( handle == _comm )
    {
        // Akutuelle Baudrate und Timeout anpassen.
        _value = baudRate;
        _timeout = timeout;
    }
#endif

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schliessen der seriellen Kommunikationsschnittstelle.
 */
int Channel4Base::closeSerial()
{
    int res = IV_ERROR_SUCCESS;
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );
    
#ifdef _MSC_VER
    try
    {
        if ( !CancelIo( _comm ) )
        {
            trace( "%s(%i):ERROR(1): %s() [ CancelIo fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }

    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( _comm, CLRRTS ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ EscapeCommFunction fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }
    if ( (res == IV_ERROR_SUCCESS)  && !EscapeCommFunction( _comm, CLRDTR ) )
    {
        trace( "%s(%i):ERROR(1): %s() [ EscapeCommFunction fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        res = IV_ERROR_FAIL;
    }

    try
    {
        if ( !CloseHandle( _comm ) )
        {
            trace( "%s(%i):ERROR(1): %s() [ CloseHandle fail (%i) ]\n", __FILE__, __LINE__, __func__, GetLastError() );
        }
    }
    catch ( ... )
    {
    }
#else
    int rc;
    int termio_cm;

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( _comm, TIOCMGET, &termio_cm )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    if ( (termio_cm & TIOCM_RTS) )
        termio_cm &= ~TIOCM_RTS;

    if ( (termio_cm & TIOCM_DTR) )
        termio_cm &= ~TIOCM_DTR;

    if ( (res == IV_ERROR_SUCCESS) && (rc = ioctl( _comm, TIOCMSET, &termio_cm )) < 0 )
    {
        trace( "%s(%i):ERROR(1): %s() [ ioctl fail ]\n", __FILE__, __LINE__, __func__ );
        res = IV_ERROR_FAIL;
    }

    ::close( _comm );
#endif
    _comm = INVALID_HANDLE_VALUE;

    return res;
}

// ##################################################################
/**
 * @brief Lesen aus der seriellen Kommunikationsschnittstelle.
 */
int Channel4Base::readSerial( void *buffer, int readCount, int &readInCount, int readTrials )
{
    trace("%s(%i):INFO(3): %s( %p, %i, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials);

    int res;
    unsigned char trials = 0;
    readInCount = 0;
#ifndef _MSC_VER
    struct timeval seltime;
    fd_set readfds;
    errno = 0;
#endif

    // Daten emfpangen
    for ( ; (readCount != readInCount) && (trials < readTrials); ++trials )
    {
        res = 0;
#ifdef _MSC_VER
        if ( !ReadFile( _comm, (unsigned char*)buffer + readInCount, readCount - readInCount, (unsigned long*)&res, NULL ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%i):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %i ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }

            trace( "%s(%i):ERROR(1): %s() [ trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
            return -IV_ERROR_RECEIVE;
        }
#else
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( _comm, (fd_set FAR *) &readfds );
        seltime.tv_sec  = _timeout / 1000;
        seltime.tv_usec = (_timeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (struct timeval *)&seltime )) < 0 )
        {
            trace( "%s(%i):ERROR(1): %s() [ trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        { 
            if ( (res = (int)read( _comm, (unsigned char*)buffer + readInCount, readCount - readInCount )) < 0 )
            {
                trace( "%s(%i):ERROR(1): %s() [ trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
                return -IV_ERROR_RECEIVE;
            }
        }
#endif
        if ( res == 0 )
        {
#ifdef _MSC_VER
            trace( "%s(%i):WARNING(1): %s() [ Leseversuch ohne Daten. (WinError: %i)]\n", __FILE__, __LINE__, __func__, GetLastError() );
#else
            trace( "%s(%i):WARNING(1): %s() [ Leseversuch ohne Daten. ]\n", __FILE__, __LINE__, __func__ );
#endif
            break;
        }
        else if ( res < (int)(readCount - readInCount) )
        {
            trace( "%s(%i):WARNING(1): %s() [ Nachlesen erforderlich. trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, __func__, trials, res, readCount - readInCount, readInCount );
        }

        readInCount += res;
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        _amountSize += res;
    }
    
    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( readCount != readInCount )
    {
        trace( "%s(%i):ERROR(1): %s() [ trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, trials, readCount, readInCount );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schreiben in die serielle Kommunikationsschnittstelle.
 */
int Channel4Base::writeSerial( const void *buffer, int writeCount, int &writeOutCount, int writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

#ifndef _MSC_VER
    errno = 0;
#endif

    int res = IV_ERROR_SUCCESS;
    unsigned char trials = 0;
    writeOutCount = 0;

    for ( ; (writeCount != writeOutCount) && (trials < writeTrials); ++trials )
    {
#ifdef _MSC_VER
        if ( !WriteFile( _comm, (unsigned char*)buffer + writeOutCount, writeCount - writeOutCount, (unsigned long*)&res, NULL ) )
        {
            if ( GetLastError() == ERROR_FILE_NOT_FOUND )
            {
                trace( "%s(%i):ERROR(1): %s() [ Schnittstelle wird geschlossen, error: %i ]\n", __FILE__, __LINE__, __func__, GetLastError() );
                close();
            }
#else
            
        if ( (res = write( _comm, (unsigned char*)buffer + writeOutCount, writeCount - writeOutCount )) < 0 )
        {
#endif
#ifdef _MSC_VER
            trace( "%s(%i):ERROR(1): %s() [ trial: %i, writeCount: %i, writeOutCount: %i, error: %i ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount, GetLastError() );
            if ( GetLastError() == ERROR_ALREADY_EXISTS )
            {            
                // 2014-04-03
                // Workaround um bei USB-COM-Ports ein blockieren zu verhindern.
                Sleep( 3000 );
            }
#else
            trace( "%s(%i):ERROR(1): %s() [ trial: %i, writeCount: %i, writeOutCount: %i ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount );
#endif
            return -IV_ERROR_TRANSMIT;
        }

        // Wurden alle Zeichen versendet?
        if ( res == 0 )
        {
            trace( "%s(%i):WARNING(1): %s() [ Es wurden keine Daten gesendet. ]\n", __FILE__, __LINE__, __func__ );
            break;

        }
        else if ( res < (writeCount - writeOutCount) )
        {		
            trace( "%s(%i):WARNING(1): %s() [ Nachsenden erforderlich. trial: %i, waren %i, gefordert %i, bereits %i]\n", __FILE__, __LINE__, __func__, trials, res, writeCount - writeOutCount, writeOutCount );
        }

        writeOutCount += res;
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        _writtenSize += res;
    }

    // Wurden nicht alle Zeichen versendet?
    if ( writeCount != writeOutCount )
    {
        trace( "%s(%i):ERROR(1): %s() [ trial: %i, writeCount: %i, writeOutCount: %i ]\n", __FILE__, __LINE__, __func__, trials, writeCount, writeOutCount );
        return -IV_ERROR_TRANSMIT;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Oeffnen einer TCP/IP Kommunikationsschnittstelle.
 */
int Channel4Base::openSocket( const std::string &host, int port, int timeout )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i )\n", __FILE__, __LINE__, __func__, host.c_str(), port, timeout );

    SOCKADDR_IN sin;
    memset( &sin, 0, sizeof(sin) );
    sin.sin_family = AF_INET;
    sin.sin_port = htons( port );
    sin.sin_addr.s_addr = inet_addr( host.data() );

    // Kanal bereits belegt?
    if ( getType() != Undefined )
        return IV_ERROR_FAIL;

    if ( sin.sin_addr.s_addr == INADDR_NONE )
    { 
        // Adresse ueber DNS aufloesen
        HOSTENT *he;
        he = gethostbyname( host.data() );
        if ( !he )
        {
            trace( "%s(%i):ERROR(1): %s() [ gethostbyname fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
            return IV_ERROR_FAIL;
        }

        sin.sin_addr.s_addr = *((unsigned int*)he->h_addr);
    }

    if ( (_socket = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET )
    {
        trace( "%s(%i):ERROR(1): %s() [ socket fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
      
    if ( connect( _socket, (SOCKADDR*)&sin, sizeof(sin) ) == SOCKET_ERROR )
    {
        trace( "%s(%i):ERROR(1): %s() [ connect fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        closeSocket();
        return IV_ERROR_FAIL;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Oeffnen einer TCP/IP Kommunikationsschnittstelle.
 */
int Channel4Base::openSocketServer( const std::string &host, int port, int timeout )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i )\n", __FILE__, __LINE__, __func__, host.c_str(), port, timeout );

    // Kanal bereits belegt?
    if ( getType() != Undefined )
        return IV_ERROR_FAIL;

    // Nicht bereits aktiv?
    if ( _communicationServerThread != 0 )
        return IV_ERROR_FAIL;

    // Struktur für Listensocket aufbauen.
    memset( &_serverThreadData.sin, 0, sizeof(SOCKADDR_IN) );
    _serverThreadData.sin.sin_family = AF_INET;
    _serverThreadData.sin.sin_port = htons( port );
    _serverThreadData.sin.sin_addr.s_addr = inet_addr( host.data() );

    if ( _serverThreadData.sin.sin_addr.s_addr == INADDR_NONE )
    { 
        // Adresse ueber DNS aufloesen
        HOSTENT *he;
        he = gethostbyname( host.data() );
        if ( !he )
        {
            trace( "%s(%i):ERROR(1): %s() [ gethostbyname fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
            return IV_ERROR_FAIL;
        }

        _serverThreadData.sin.sin_addr.s_addr = *((unsigned int*)he->h_addr);
    }

    _serverThreadData.timeout = timeout;

    // Für den verbundenen Socket.
    memset( &_serverThreadData.sa, 0, sizeof(SOCKADDR_IN) );

    // Verbindungstyp einstellen
    _type = Channel4Base::SocketServer;

    // Hauptthread starten; Dieser besorgt die Verbindunsannahme und Befehlsabarbeitung.
    int i;
#ifdef _MSC_VER
    DWORD threadID;
#endif
    _initSignal = -1; // Initialisierungssignal auf undefiniert setzen
#ifdef _MSC_VER
    _communicationServerThread = ::CreateThread( NULL, 0, communicationServerThread, this, 0, &threadID );
#else
    pthread_create( &_communicationServerThread, NULL, communicationServerThread, this );
#endif

    // Hochlauf des Workers pruefen
    for ( i = 0; (_communicationServerThread != 0) && (_initSignal == -1) && (i < 100); ++i )
        Sleep( 100 );

    // Wurde ein Fehler ausgeloest.
    if ( _initSignal != IV_ERROR_SUCCESS )
    {
        // Grundlegender Initialiserungsfehler?
        if ( _initSignal == -1 )
            return IV_ERROR_FAIL;
        
        // Fehler von Workerthread 1:1 weitergeben.
        return _initSignal;
    }

    return IV_ERROR_SUCCESS;
}

#ifdef _MSC_VER
unsigned long WINAPI Channel4Base::communicationServerThread( LPVOID lpParam )
#else
void *Channel4Base::communicationServerThread( void *lpParam )
#endif
{
    // Muss einen Parameter uebergeben bekommen vom Type DatafoxLibraryIV.
    if ( lpParam == NULL )
        return 0;

    // Referenz auf den uebergebenen Parameter
    Channel4Base &own = *reinterpret_cast<Channel4Base *>( lpParam );

#ifdef _MSC_VER
#ifndef CONST
#define CONST const
#endif
#else
#define CONST
#endif

    // Hauptschleife ausfuehren?
    int run = FALSE;

    int rc;
    struct timeval seltime;
    fd_set readfds;

    // Zustandsstatus fuer Listen - Socket
    own._communicationServerThreadState = Down;

    // Timeout fuer Verbindungsanfragen. 
    //Achtung: Nicht auf 0 da sonst der Thread blockiert wuerde!
    seltime.tv_sec  = 1;
    seltime.tv_usec = 0;

    // Sockethandle reservieren
    own._listenSocket = socket( AF_INET, SOCK_STREAM, 0 );
    if ( own._listenSocket != INVALID_SOCKET ) 
    {
        // Socket mit lokalem Port verbinden.
        if ( ::bind( own._listenSocket, reinterpret_cast<SOCKADDR *>( &own._serverThreadData.sin ), sizeof( SOCKADDR_IN ) ) == SOCKET_ERROR )
        {
#ifdef _MSC_VER
            if ( WSAGetLastError() == WSAEADDRINUSE )
            {
#else
            if ( WSAGetLastError() == EADDRINUSE )
            {
#endif
                own.trace( "%s(%i):ERROR(1): %s [ bind fail (%i) Die gewaehlte Portnummer ist in Gebrauch (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError(), own._serverThreadData.sin.sin_port );
                // Initialisierungsfehler setzten
                own._initSignal = IV_ERROR_AC_PORTINUSE;
            }
            else
            {
                own.trace( "%s(%i):ERROR(1): %s [ bind fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                // Initialisierungsfehler setzten
                own._initSignal = IV_ERROR_AC_LISTENER;
            }

#ifdef _MSC_VER
            if ( closesocket( own._listenSocket ) == SOCKET_ERROR )
            {
                own.trace( "%s(%i):ERROR(1): %s [ closesocket fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                // Initialisierungsfehler setzten
                own._initSignal = IV_ERROR_AC_LISTENER;
            }
#else
            if ( shutdown(own._listenSocket, 2 ) == SOCKET_ERROR )
            {
                own.trace( "%s(%i):ERROR(1): %s [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                // Initialisierungsfehler setzten
                own._initSignal = IV_ERROR_AC_LISTENER;
            }
            if ( ::close( own._listenSocket ) == SOCKET_ERROR )
            {
                own.trace( "%s(%i):ERROR(1): %s [ close fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                // Initialisierungsfehler setzten
                own._initSignal = IV_ERROR_AC_LISTENER;
            }

#endif
            own._listenSocket = INVALID_SOCKET;
        }
        else
        {
            // Mit Reset - Status beginnen
            run = TRUE;
            own._communicationServerThreadState = Reset;
            // Initialisierung erfolgreich
            own._initSignal = IV_ERROR_SUCCESS;
        }
    }
    else
    {
        own.trace( "%s(%i):ERROR(1): %s [ socket fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        // Initialisierungsfehler setzen
        own._initSignal = IV_ERROR_AC_LISTENER;
    }

    // Hauptschleife des Worker
    while ( run )
    {
        // Beendigungssignal abfragen
        if ( own._endSignal )
        {
            // Verbindung herunterfahren
            if (own._socket != INVALID_SOCKET)
            {
                // Senden am Socket verhindern
                if (shutdown(own._socket, 0x01/*SD_SEND*/) == SOCKET_ERROR)
                {
                    own.trace("%s(%i):ERROR(1): %s [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                }
                else 
                {
                    // Alle anstehenden Daten lesen
                    do {

                        FD_ZERO((fd_set FAR *) &readfds);
                        FD_SET(own._socket, (fd_set FAR *) &readfds);

                        if ((rc = select(FD_SETSIZE,
                            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                            (CONST struct timeval FAR *)&seltime)) == SOCKET_ERROR)
                        {
                            own.trace("%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                        }
                        else if (rc)
                        {
                            char packet[0x300];
                            // Anstehende Daten lesen
                            if ((rc = recv(own._socket, packet, sizeof(packet), 0 )) == SOCKET_ERROR)
                            {
                                own.trace("%s(%i):ERROR(1): %s [ recv fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());							
                            }
                            else if (rc == 0)
                            {
                                // Disconnect von Gegenstelle
                            }
                        }
                    } while((rc > 0) && (rc != SOCKET_ERROR));

                    // Verbindung schliessen
                    own.closeSocket();
                }

                own._type = Channel4Base::Undefined;
            }

            // Listensocket herunterfahren
            if ( own._listenSocket != INVALID_SOCKET )
            {
                if ( own._communicationServerThreadState != Down )
                {
                    // Anstehende Verbindungsgesuche verwerfen
                    if ( listen( own._listenSocket, 1 ) == SOCKET_ERROR )
                    { 
                        own.trace("%s(%i):ERROR(1): %s [ listen fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                    }
                }
#ifdef _MSC_VER
                // Listensocket schliessen
                if ( closesocket( own._listenSocket ) == SOCKET_ERROR )
                {
                    own.trace("%s(%i):ERROR(1): %s [ closesocket fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                }
#else
                if ( shutdown( own._listenSocket, 2 ) == SOCKET_ERROR )
                {
                    own.trace( "%s(%i):ERROR(1): %s [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                }
                if ( ::close( own._listenSocket ) == SOCKET_ERROR )
                {
                    own.trace( "%s(%i):ERROR(1): %s [ close fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                }
#endif
                own._listenSocket = INVALID_SOCKET;

                // Eventuelle Verbindungen beenden
                own.closeSocket();
            }

            // Zustand setzen
            own._communicationServerThreadState = Down;

            // Hauptschleife verlassen
            run = FALSE;
            continue;
        }

        // Je nach Zustand den Listen-Socket verwalten
        switch ( own._communicationServerThreadState )
        {
        case Down: 
            // Aufgrund eines Fehlers abgeschaltet
            break;

        case Reset: // Verwirft eventuell anstehende Verbindungsanforderungen,
            // durch (oder) Erzeugung Listensocket.
            if ( listen( own._listenSocket, 1 ) == SOCKET_ERROR )
            { 
                own.trace( "%s(%i):ERROR(1): %s [ listen fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );

                own._communicationServerThreadState = Down;
                own.trace( "%s(%i):INFO(1): %s [ State change to down... ]\n", __FILE__, __LINE__, __func__ );
            } else {

                own._communicationServerThreadState = Listening;
                own.trace( "%s(%i):INFO(1): %s [ State change to listening... ]\n", __FILE__, __LINE__, __func__ );
            }
            break;

        case Listening: // Wartet auf connect. Im "listening" - Zustand.

            FD_ZERO( reinterpret_cast<fd_set FAR *>( &readfds ) );
            FD_SET( own._listenSocket, reinterpret_cast<fd_set FAR *>( &readfds ) );

            if ( (rc = select( FD_SETSIZE,
                reinterpret_cast<fd_set FAR *>( &readfds ), reinterpret_cast<fd_set FAR *>( 0 ), reinterpret_cast<fd_set FAR *>( 0 ),
                &seltime ) ) == SOCKET_ERROR )
            {
                own.trace( "%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
            }
            else if ( rc )
            {
                // Verbindungsanforderung auf Listensocket im naechsten
                // Druchlauf verarbeiten.
                own._communicationServerThreadState = Accepting;
                own.trace( "%s(%i):INFO(1): %s [ State change to accepting... ]\n", __FILE__, __LINE__, __func__ );
            } 
            break;

        case Accepting: // Verbindungsanforderung verarbeiten
        { 
            // Verbindungsgesuch annehmen.
            int len = sizeof(own._serverThreadData.sa);
            own._socket = accept(own._listenSocket, (SOCKADDR *)&own._serverThreadData.sa, (socklen_t*)&len);

            if(own._socket == INVALID_SOCKET)
            {
                own.trace("%s(%i):ERROR(1): %s [ accept fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());
                // Initialisierungsfehler setzen.
                own._initSignal = IV_ERROR_FAIL;
                // Fehlerfall abfangen
                own._communicationServerThreadState = Reset;
                own.trace( "%s(%i):INFO(1): %s [ State change to reset... ]\n", __FILE__, __LINE__, __func__ );
                continue;
            }

            // Verbindungsmeldung ausgeben
            own.trace("%s(%i):INFO(1): %s [ (%i) Connected to %s, port %d ]\n", __FILE__, __LINE__, __func__, own.getId(), inet_ntoa(own._serverThreadData.sa.sin_addr), htons(own._serverThreadData.sa.sin_port));

            // Mindesttimeout sicherstellen
            if (own.getTimeout() < 3000)
            {
                own.setTimeout(3000);
            }

            // Verbindungsanfrage verarbeitet
            own._communicationServerThreadState = Connected;
            own.trace( "%s(%i):INFO(1): %s [ State change to connected... ]\n", __FILE__, __LINE__, __func__ );
            break;
        }

        case Connected:
        {
            if ( own.callServer() != IV_ERROR_SUCCESS )
            {
                // Verbindung schliessen
                own.closeSocket();
                own.trace("%s(%i):INFO(1): %s [ Disconnect ]\n", __FILE__, __LINE__, __func__, own.getId(), inet_ntoa(own._serverThreadData.sa.sin_addr), htons(own._serverThreadData.sa.sin_port));

                own._communicationServerThreadState = Reset;
                own.trace( "%s(%i):INFO(1): %s [ State change to reset... ]\n", __FILE__, __LINE__, __func__ );
            }
            break;
        }
        }

        // Kurze Ausszeit um die Systemlast zu reduzieren.
        Sleep( 10 );
    }

    // Threadzeiger ruecksetzen
    own._communicationServerThread = 0;

    // Thread Beendigung ruecksetzen
    own._endSignal = FALSE;

#ifndef _MSC_VER
    pthread_exit( NULL );
#endif
    return 0;
}

// ##################################################################
/**
* @brief Beendigungssignal im Workerthread setzen und auf dessen beendigung warten.
*/
void Channel4Base::closeSocketServer()
{
    trace("%s(%i):INFO(3): %s", __FILE__, __LINE__, __func__);

    // Socket-Thread deaktivieren, wenn aktiv
    if ( _communicationServerThread != 0 )
    {
        // WorkerThread fuer Socket beenden.
        _endSignal = TRUE;

        // Thread umschalten
        Sleep(10);

        // Thread Beendigung abwarten (communicationThread stellt _endSignal auf FALSE)
        int i;
        for (i=0; i < (getTimeout()*6 / 100) && _endSignal; i++)
        {				
            Sleep(100);
        }

        // Hartes Schliessen, wenn noch nicht selbst beendet.
        if (_communicationServerThread)
        {
#ifdef _MSC_VER
            ::CloseHandle(_communicationServerThread);
#else
            pthread_exit(&_communicationServerThread);
#endif
            _communicationServerThread = 0;
        }

        // Kanaltyp ist wieder frei.
        _type = Channel4Base::Undefined;
    }
}


// ##################################################################
/**
 * @brief Schliessen der TCP/IP Kommunikationsschnittstelle.
 */
int Channel4Base::closeSocket()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#ifdef _MSC_VER
    if ( closesocket( _socket ) == SOCKET_ERROR )
    {
        trace( "%s(%i):ERROR(1): %s() [ closesocket fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#else
    if ( shutdown( _socket, 2 ) == SOCKET_ERROR )
    {
        trace( "%s(%i):ERROR(1): %s() [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
    }
    if ( ::close( _socket ) == SOCKET_ERROR )
    {
        trace( "%s(%i):ERROR(1): %s() [ close fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        return IV_ERROR_FAIL;
    }
#endif

    _socket = INVALID_SOCKET;

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Lesen aus der TCP/IP Kommunikationsschnittstelle.
 */
int Channel4Base::readSocket(void *buffer, int readCount, int &readInCount, int readTrials)
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials );

#if _MSC_VER
#define MYCONST const
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    int res;
    unsigned char trials = 0;
    unsigned char lTrial = readCount;

    readInCount = 0;

    // Daten emfpangen
    for ( ; (readCount != readInCount) && (trials < readTrials); ++trials )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( _socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = _timeout / 1000;
        seltime.tv_usec = (_timeout % 1000) * 1000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, readCount: %i, readInCount: %i, error: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, readCount, readInCount, WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {			
            if ( (res = recv( _socket, (char*)buffer + readInCount, readCount - readInCount, 0 )) == SOCKET_ERROR )
            {
#ifdef _MSC_VER
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%i):ERROR(1): %s() [ id: %i, recv error: %i ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, readCount: %i, readInCount: %i, error: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, readCount, readInCount, WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, readCount, readInCount );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
            else if ( res < (readCount - readInCount) )
            {
                trace( "%s(%i):WARNING(1): %s() [ Nachlesen erforderlich. id: %i, trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, res, readCount - readInCount, readInCount );
                // Fuer Kommunikation ueber Com-Server mit Mehrfachdocking viele Teilpackete
                if ( lTrial )
                    trials--;
                else
                    lTrial = (readCount - readInCount);

                lTrial--;
            }
        }
        else
        {
            trace( "%s(%i):WARNING(1): %s() [ Es wurden keine Daten gelesen. id: %i ]\n", __FILE__, __LINE__, __func__, getId() );
            break;
        }

        readInCount += res;
        // Gesamtanzahl gelesener Bytes fuer Statistik halten
        _amountSize += res;
    }

    // Wurde gewuenschte Anzahl nicht empfangen?
    if ( readCount != readInCount )
    {
        trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, readCount: %i, readInCount: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, readCount, readInCount );
        return -IV_ERROR_RECEIVE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schreiben in die TCP/IP Kommunikationsschnittstelle.
 */
int Channel4Base::writeSocket( const void *buffer, int writeCount, int &writeOutCount, int writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

    int res;
    unsigned char trials = 0;

    writeOutCount = 0;

    for ( ; (writeCount != writeOutCount) && (trials < writeTrials); ++trials )
    {
        // Daten senden
        if ( (res = send( _socket, (char*)buffer + writeOutCount, writeCount - writeOutCount, 0 )) == SOCKET_ERROR )
        {
#ifdef _MSC_VER
            if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
            {
                trace( "%s(%i):ERROR(1): %s) [ id: %i, send error: %i ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
#endif
            trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, writeCount %i, writeOutCount: %i, Error: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, writeCount, writeOutCount, WSAGetLastError() );
            return -IV_ERROR_TRANSMIT;
        }
        
        // Wurden alle Daten versendet?
        if ( res == 0 )
        {
            trace( "%s(%i):WARNING(1): %s() [ Es wurden keine Daten gesendet. id: %i ]\n", __FILE__, __LINE__, __func__, getId() );
            break;
        }
        else if ( res < (writeCount - writeOutCount) )
        {		
            trace( "%s(%i):WARNING(1): %s() [ Nachsenden erforderlich. id: %i, trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, res, writeCount - writeOutCount, writeOutCount );
        }

        writeOutCount += res;
        // Gesamtanzahl geschriebener Bytes fuer Statistik halten
        _writtenSize += res;
    }

    // Wurden nicht alle Daten im vorgegebenen Rahmen versendet?
    if ( writeCount != writeOutCount )
    {
        trace( "%s(%i):ERROR(1): %s() [ id: %i, trial: %i, writeCount %i, writeOutCount: %i ]\n", __FILE__, __LINE__, __func__, getId(), trials, writeCount, writeOutCount );
        return -IV_ERROR_TRANSMIT;
    }

    // Ausfuehrung ist erfolgt.
    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Verwerfen der Daten an der TCP/IP Kommunikationsschnittstelle.
*/
int Channel4Base::flushSocket()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

#if _MSC_VER
#define MYCONST const
#else
#define MYCONST
#endif
    struct timeval seltime;
    fd_set readfds;

    int res;
    char buffer[1024];

    // Daten emfpangen
    for ( ;; )
    {
        FD_ZERO( (fd_set FAR *) &readfds );
        FD_SET( _socket, (fd_set FAR *) &readfds );
        seltime.tv_sec  = 0;
        seltime.tv_usec = 500000;

        if ( (res = select( FD_SETSIZE,
            (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
            (MYCONST struct timeval FAR *)&seltime )) < 0 )
        {
            trace( "%s(%i):ERROR(1): %s() [ id: %i, error: %i ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
            return -IV_ERROR_RECEIVE;
        }

        // Liegen Daten an?
        if ( res )
        {			
            if ( (res = recv( _socket, buffer, sizeof(buffer), 0 )) == SOCKET_ERROR )
            {
#ifdef _MSC_VER
                if ( WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError() )
                {
                    trace( "%s(%i):ERROR(1): %s() [ id: %i, recv error: %i ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
                    return -IV_ERROR_CHANNEL_OPEN_REMOTE;
                }
#endif
                trace( "%s(%i):ERROR(1): %s() [ id: %i, error: %i ]\n", __FILE__, __LINE__, __func__, getId(), WSAGetLastError() );
                return -IV_ERROR_RECEIVE;
            }
            else if ( res == 0 )
            {
                // Wenn Daten gemeldet werden,
                // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                // Verbindung wurde von Gegenseite geschlossen.
                trace( "%s(%i):ERROR(1): %s() [ id: %i ]\n", __FILE__, __LINE__, __func__, getId() );	
                return -IV_ERROR_CHANNEL_OPEN_REMOTE;
            }
        }
        else
        {
            break;
        }
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Lesen aus der Kommunikationsschnittstelle.
 */
int Channel4Base::readData( void *buffer, int readCount, int &readInCount, int readTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, readCount, readInCount, readTrials );

    int res;
    switch ( getType() )
    {
        case Serial:
            res = readSerial( buffer, readCount, readInCount, readTrials );
            break;

        case Socket:
        case SocketServer:
            res = readSocket( buffer, readCount, readInCount, readTrials );
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, __func__, getType() );
            return -IV_ERROR_RECEIVE;
    }

    return res;
}

// ##################################################################
/**
 * @brief Schreiben in die Kommunikationsschnittstelle.
 */
int Channel4Base::writeData( const void *buffer, int writeCount, int &writeOutCount, int writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, %i, %i )\n", __FILE__, __LINE__, __func__, buffer, writeCount, writeOutCount, writeTrials );

    int res;
    switch ( getType() )
    {
        case Serial:
            res = writeSerial( buffer, writeCount, writeOutCount, writeTrials );
            break;

        case Socket:
        case SocketServer:
            res = writeSocket( buffer, writeCount, writeOutCount, writeTrials );
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, __func__, getType() );
            return -IV_ERROR_TRANSMIT;
    }

    return res;
}

